import chalk from 'chalk';
import chokidar from 'chokidar';
import cron from 'cron';
import notifier from 'node-notifier';
import nodemailer from 'nodemailer';
import fs from 'fs';
import fspath from 'path';
import { exec } from 'child_process';
import { spawn } from 'child_process';
import httpRequest from 'request';
import scriptkit from './spellu-scriptkit.mjs';
import SON from './grammar-son.mjs';
var influe;
(function (influe) {
    const packageConfigPath = fspath.join(fspath.dirname(new URL(import.meta.url).pathname), '../package.json');
    const packageConfig = SON.parse(fs.readFileSync(packageConfigPath, { encoding: 'utf8' })) || {};
    influe.metadata = {
        name: 'influe',
        version: packageConfig.version || '',
        // config: packageConfig,
    };
})(influe || (influe = {}));
(function (influe) {
    function wrapInArray(object) {
        return object === undefined ? [] : (Array.isArray(object) ? object : [object]);
    }
    influe.wrapInArray = wrapInArray;
    function mapObject(object, converter) {
        const result = {};
        for (const key in object) {
            result[key] = converter(object[key]);
        }
        return result;
    }
    influe.mapObject = mapObject;
    // export function tap<T extends {}>(
    // 	object: T
    // ): T {
    // 	return object
    // }
})(influe || (influe = {}));
(function (influe) {
    let Trigger;
    (function (Trigger) {
        Trigger["Add"] = "add";
        Trigger["Change"] = "change";
        Trigger["Unlink"] = "unlink";
        Trigger["AddDir"] = "addDir";
        Trigger["UnlinkDir"] = "unlinkDir";
    })(Trigger = influe.Trigger || (influe.Trigger = {}));
})(influe || (influe = {}));
(function (influe) {
    let TouchPoint;
    (function (TouchPoint) {
        TouchPoint["Make"] = "Make";
        TouchPoint["Watch"] = "Watch";
        TouchPoint["Run"] = "Run";
        TouchPoint["Test"] = "Test";
        TouchPoint["Trace"] = "Trace";
        TouchPoint["Config"] = "Config";
        TouchPoint["Task"] = "Task";
    })(TouchPoint = influe.TouchPoint || (influe.TouchPoint = {}));
    function touch(point, timing, ...args) {
        switch (point) {
            case TouchPoint.Make:
            case TouchPoint.Watch:
            case TouchPoint.Run:
            case TouchPoint.Test:
            case TouchPoint.Trace:
            case TouchPoint.Config:
            case TouchPoint.Task:
                if (timing == 'pre') {
                    console.log(chalk.magenta(influe.metadata.name), chalk.yellow(point), ...args);
                }
                else if (timing == 'error') {
                    const error = args.shift();
                    console.error(chalk.magenta(influe.metadata.name), chalk.yellow(point), chalk.red('ERROR'), error.message, ...args);
                }
                else if (timing == 'log') {
                    console.log(chalk.magenta(influe.metadata.name), chalk.yellow(point), ...args);
                }
                else {
                    throw new Error("unknown timing");
                }
                break;
        }
    }
    influe.touch = touch;
})(influe || (influe = {}));
(function (influe) {
    const defaults = {
        debug: true,
        path: undefined,
        data: { default: {} },
        tasks: [],
        files: [],
        events: [],
    };
    let ConfigSyntax;
    (function (ConfigSyntax) {
        ConfigSyntax[ConfigSyntax["SON"] = 0] = "SON";
        ConfigSyntax[ConfigSyntax["JSON"] = 1] = "JSON";
    })(ConfigSyntax || (ConfigSyntax = {}));
    function configure(path = './influe.config') {
        let syntax;
        if (fs.existsSync(path)) {
            if (fs.statSync(path).isDirectory()) {
                path = path.replace(/(\/+)$/, '') + '/influe.config';
            }
        }
        else {
            influe.touch(influe.TouchPoint.Config, 'log', 'cwd', process.cwd());
            throw new Error(`config file is not found. "${path}"`);
        }
        path = fspath.resolve(path);
        // 構成をロードする
        console.info(chalk.magenta(influe.metadata.name), chalk.yellow('Config'), `"${path}"`);
        const source = SON.parse(fs.readFileSync(path, 'utf8'));
        return loadConfig(path, source);
    }
    influe.configure = configure;
    function loadConfig(path, source) {
        source = { debug: false, ...source };
        // TODO: check
        return {
            ...defaults,
            debug: source.debug,
            path,
            data: source.data || {},
            events: influe.wrapInArray(source.event)
                .map((_, index) => loadEventSection(index, _)),
            files: influe.wrapInArray(source.file)
                .map((_, index) => loadFileSection(index, _)),
            tasks: influe.wrapInArray(source.task)
                .map((_, index) => loadTaskSection(index, _)),
        };
    }
    function loadEventSection(index, _) {
        // TODO: check
        const id = _.id || `E#${index + 1}`;
        if (_.time) {
            return loadTimeEvent(id, _);
        }
        else if (_.file) {
            return loadFileEvent(id, _);
        }
        else if (_.key) {
            return loadKeyEvent(id, _);
        }
        else {
            throw new Error(`${id}: must "time" or "file" or "keyboard".`);
        }
    }
    function loadTimeEvent(id, _) {
        return {
            id,
            type: 'time:cron',
            crontime: _.time.cron,
            timezone: _.time.zone || '',
            makeTargets: influe.wrapInArray(_.make).map(_ => loadEventMake(_)),
            runTargets: influe.wrapInArray(_.run),
            enabled: _.enabled === undefined ? true : _.enabled,
        };
    }
    function loadFileEvent(id, _) {
        return {
            id,
            type: 'file',
            paths: influe.wrapInArray(_.file.path),
            triggers: influe.wrapInArray(_.file.trigger),
            makeTargets: influe.wrapInArray(_.make).map(_ => loadEventMake(_)),
            runTargets: influe.wrapInArray(_.run),
            enabled: _.enabled === undefined ? true : _.enabled,
        };
    }
    function loadKeyEvent(id, _) {
        return {
            id,
            type: 'key',
            key: _.key,
            makeTargets: influe.wrapInArray(_.make).map(_ => loadEventMake(_)),
            runTargets: influe.wrapInArray(_.run),
            enabled: _.enabled === undefined ? true : _.enabled,
        };
    }
    function loadEventMake(_) {
        if (_.state === undefined)
            throw new Error('state must value.');
        return {
            files: influe.wrapInArray(_.file),
            state: _.state
        };
    }
    function loadFileSection(index, _) {
        // TODO: check
        return {
            id: _.id || `F#${index + 1}`,
            paths: influe.wrapInArray(_.path),
            states: influe.mapObject(_.state || {}, _ => influe.wrapInArray(_).map(_ => loadFileState(_))),
        };
    }
    function loadFileState(_) {
        if (typeof _ === 'string') {
            return _;
        }
        else {
            return {
                id: '_',
                action: _.action || 'command',
                arguments: _.argument || {},
            };
        }
    }
    function loadTaskSection(index, _) {
        // TODO: check
        return {
            id: _.id || `T#${index + 1}`,
            action: _.action || 'command',
            arguments: _.argument || {},
        };
    }
})(influe || (influe = {}));
(function (influe) {
    var command;
    (function (command) {
        function make(config, state) {
            const files = config.files.filter(_ => state in _.states);
            influe.touch(influe.TouchPoint.Make, 'pre', chalk.grey('state:') + chalk.underline(state), chalk.grey('file:') + chalk.underline(files.length.toString()));
            for (const file of files) {
                influe.updateFile(config, file, state);
            }
        }
        command.make = make;
    })(command = influe.command || (influe.command = {}));
})(influe || (influe = {}));
(function (influe) {
    var command;
    (function (command) {
        function run(config, taskId) {
            influe.touch(influe.TouchPoint.Run, 'pre', taskId);
            const task = config.tasks.find(_ => _.id == taskId);
            if (task === undefined) {
                throw new Error(`Task "${taskId}" is not found.`);
            }
            influe.runTask(task.id, task, { data: config.data });
        }
        command.run = run;
    })(command = influe.command || (influe.command = {}));
})(influe || (influe = {}));
(function (influe) {
    var command;
    (function (command) {
        function watch(config, configPath) {
            influe.touch(influe.TouchPoint.Watch, 'pre');
            // イベントの準備
            influe.startTimeWatch(config);
            influe.startFileWatch(config);
            influe.startConsoleWatch(config);
            // 設定ファイルの監視
            if (config.path) {
                const watcher = chokidar.watch(config.path);
                watcher
                    .on('change', path => {
                    console.info('構成ファイル変更検知');
                    console.info('停止');
                    influe.stopTimeWatch();
                    influe.stopFileWatch();
                    influe.stopConsoleWatch();
                    console.info('再開');
                    config = influe.configure(configPath);
                    influe.startTimeWatch(config);
                    influe.startFileWatch(config);
                    influe.startConsoleWatch(config);
                });
            }
            console.info('開始');
            // プロセス終了イベントのリスニング
            process.on('exit', () => {
                console.log('[Process:Exited]');
            });
        }
        command.watch = watch;
        function stopWatchAndExit() {
            console.log();
            console.info('停止');
            process.exit();
        }
        command.stopWatchAndExit = stopWatchAndExit;
    })(command = influe.command || (influe.command = {}));
})(influe || (influe = {}));
(function (influe) {
    var command;
    (function (command) {
        function test(config) {
            influe.touch(influe.TouchPoint.Test, 'pre');
        }
        command.test = test;
    })(command = influe.command || (influe.command = {}));
})(influe || (influe = {}));
(function (influe) {
    var command;
    (function (command) {
        function trace(config) {
            influe.touch(influe.TouchPoint.Trace, 'pre');
            traceFiles(config, config.files);
            traceEvents(config, config.events);
            traceTasks(config, config.tasks);
        }
        command.trace = trace;
        function traceFiles(config, files) {
            console.info(chalk.green('Files'), files.length);
            for (const file of files) {
                console.info('  ', chalk.grey('id:') + chalk.cyan.underline(file.id));
                console.info('    ', chalk.grey('location:') + chalk.blueBright(`${file.paths.length}`));
                for (const path of file.paths) {
                    console.info('      ', chalk.grey('path:') + `"${path}"`);
                }
                for (const name in file.states) {
                    console.info('    ', chalk.grey('state:') + chalk.redBright.underline(name));
                    for (const task of file.states[name]) {
                        if (typeof task === 'string') {
                            console.info('      ', chalk.grey('task:ref:') + task);
                        }
                        else {
                            console.info('      ', chalk.grey('task:action:') + chalk.yellow.underline(task.action));
                            for (const name in task.arguments) {
                                for (const value of influe.wrapInArray(task.arguments[name])) {
                                    console.info('        ', chalk.green(name), `"${value}"`);
                                }
                            }
                        }
                    }
                }
            }
        }
        function traceEvents(config, events) {
            console.info(chalk.green('Events'), events.length);
            for (const event of events) {
                console.info('  ', chalk.grey('id:') + chalk.cyan.underline(event.id));
                if (event.type == 'time:cron') {
                    console.info('    ', chalk.grey('time:'));
                    console.info('      ', chalk.grey('cron:') + `"${event.crontime}"`);
                    console.info('      ', chalk.grey('timezone:') + `"${event.timezone}"`);
                }
                else if (event.type == 'file') {
                    console.info('    ', chalk.grey('file:') + chalk.blueBright(`${event.paths.length}`));
                    for (const pathOrGlob of event.paths) {
                        const isGlob = pathOrGlob.indexOf('*') != -1;
                        console.info('      ', chalk.grey(isGlob ? 'glob:' : 'path:') + `"${pathOrGlob}"`);
                    }
                    console.info('      ', chalk.grey('trigger:'), ...event.triggers.map(_ => `"${_}"`));
                }
                console.info('    ', chalk.grey('run:') + chalk.blueBright(`${event.runTargets.length}`));
                for (const taskId of event.runTargets) {
                    console.info('      ', chalk.grey('task:') + chalk.cyan.underline(taskId));
                }
                console.info('    ', chalk.grey('make:') + chalk.blueBright(`${event.makeTargets.length}`));
                for (const object of event.makeTargets) {
                    if (object.files.length > 0) {
                        for (const fileId of object.files) {
                            console.info('      ', chalk.grey('file:') + chalk.cyan.underline(fileId), chalk.grey('state:') + chalk.redBright.underline(object.state));
                        }
                    }
                    else {
                        console.info('      ', chalk.grey('state:') + chalk.redBright.underline(object.state));
                    }
                }
            }
        }
        function traceTasks(config, tasks) {
            console.info(chalk.green('Tasks'), tasks.length);
            for (const task of tasks) {
                console.info('  ', chalk.grey('id:') + chalk.cyan.underline(task.id), chalk.grey('action:') + chalk.yellow.underline(task.action));
                if (typeof task.arguments == 'object') {
                    for (const name in task.arguments) {
                        for (const value of influe.wrapInArray(task.arguments[name])) {
                            console.info('    ', chalk.green(name), `"${value}"`);
                        }
                    }
                }
                else {
                    for (const value of influe.wrapInArray(task.arguments)) {
                        console.info('    ', `"${value}"`);
                    }
                }
            }
        }
    })(command = influe.command || (influe.command = {}));
})(influe || (influe = {}));
(function (influe) {
    influe.actions = {};
    function runAction(action, argument, context) {
        influe.actions[action](context, argument);
    }
    influe.runAction = runAction;
})(influe || (influe = {}));
(function (influe) {
    influe.actions['log'] = async (context, argument) => {
        const lines = influe.wrapInArray(argument);
        lines
            .map(_ => influe.expandTemplateText(_, context, process.env))
            .map(line => influe.touch(influe.TouchPoint.Watch, 'log', line));
    };
})(influe || (influe = {}));
(function (influe) {
    influe.actions['command'] = async (context, // == $.Dictionary
    argument) => {
        const commands = influe.wrapInArray(argument);
        influe.pauseConsoleWatch();
        const startTime = new Date().getTime();
        for (const command of commands) {
            try {
                if (!command.command || command.command.length == 0) {
                    throw new Error(`missing "command"`);
                }
                const program = influe.expandTemplateText(command.command[0], context, process.env);
                const args = command.command.slice(1)
                    .map(_ => influe.expandTemplateText(_, context, process.env));
                await influe.os.executeCommand(program, args, {
                    cwd: command.cwd ?? process.cwd(),
                    env: command.env ?? process.env,
                });
            }
            catch (error) {
                console.error(error);
            }
        }
        const endTime = new Date().getTime();
        console.log(`\u{1F52E}  Done in ${(endTime - startTime) / 1000}s.`);
        // if ()
        // console.log(`Exit Code: `)
        influe.resumeConsoleWatch();
    };
    influe.actions['shell'] = async (context, argument) => {
        const lines = influe.wrapInArray(argument);
        const commandLine = lines
            .map(line => influe.expandTemplateText(line, context, process.env))
            .join(';');
        // Note: Async call
        influe.os.executeShell(commandLine, {});
    };
    influe.actions['echo'] = async (context, argument) => {
        const lines = influe.wrapInArray(argument);
        const commandLine = lines
            .map(line => 'echo ' + influe.expandTemplateText(line, context, process.env))
            .join(';');
        influe.os.executeShell(commandLine, {});
    };
})(influe || (influe = {}));
(function (influe) {
    /**
     * generate:
     *      repository:     リポジトリ（.influeディレクトリ）のパス
     *      template:       テンプレートファイルのリポジトリからの相対パス
     *      data:           データファイルまたはデータディレクトリのリポジトリからの相対パス
     *      out:            出力ファイルのパス
     */
    influe.actions['generate'] = async (context, argument) => {
        const templatePath = fspath.join(argument.repository, argument.template);
        const dataPath = fspath.join(argument.repository, argument.data);
        const outPath = argument.out;
        //- エラーチェック
        if (!fs.existsSync(dataPath)) {
            throw new Error(`${dataPath} is not found.`);
        }
        const template = influe.os.readTextFile(templatePath);
        const data = readData(dataPath);
        // console.log(31, data, template)
        // テンプレートテキストを展開する
        const out = influe.expandTemplateText(template, context, data);
        influe.os.writeTextFile(outPath, out);
    };
    function readData(dataPath) {
        //- ディレクトリ
        if (fs.statSync(dataPath).isDirectory()) {
            return readDirectory(dataPath);
        }
        //- ファイル
        else {
            return influe.DataReader.read(dataPath);
        }
    }
    function readDirectory(dataDirectoryPath) {
        // パスリストを作成する
        const paths = fs.readdirSync(dataDirectoryPath, { encoding: 'utf8', withFileTypes: true })
            .filter(_ => _.isFile())
            .filter(_ => influe.DataReader.isSupportFile(_.name))
            .map(_ => _.name);
        // 全データを読み込む
        const repository = {};
        for (const path of paths) {
            // ファイル名から拡張子を除いた文字列を取得する
            const key = fspath.basename(path, fspath.extname(path));
            //- 1キー複数データの場合、エラーとする
            if (key in repository)
                throw new Error(`Multiple key '${key}' found.`);
            // ファイルからデータを読み込む
            repository[key] = influe.DataReader.read(fspath.join(dataDirectoryPath, path));
        }
        if ('$' in repository) {
            const data = repository['$'];
            delete repository['$'];
            for (const key in repository) {
                data[key] = repository[key];
            }
            return data;
        }
        else {
            return repository;
        }
    }
})(influe || (influe = {}));
(function (influe) {
    influe.actions['prepare'] = async (context, argument) => {
        const pathOrGlobs = influe.wrapInArray(argument.path || context.file.paths);
        const contentTemplate = argument.content || '';
        // if (fs.existsSync(path)) return
        for (const item of pathOrGlobs) {
            const path = influe.expandTemplateText(item, context, {});
            const content = influe.expandTemplateText(contentTemplate, {}, context);
            influe.os.writeTextFile(path, content);
        }
    };
    influe.actions['unlink'] = async (context, argument) => {
        const pathOrGlobs = influe.wrapInArray(argument.path || context.file.paths);
        for (const item of pathOrGlobs) {
            const path = influe.expandTemplateText(item, context, {});
            influe.os.unlinkFile(path);
        }
    };
})(influe || (influe = {}));
(function (influe) {
    influe.actions['notify'] = async (context, argument) => {
        notifier.notify({
            ...argument,
            title: argument.title,
            message: argument.message,
            // icon: argument.icon,
            sound: argument.sound,
            wait: argument.wait,
        });
    };
})(influe || (influe = {}));
(function (influe) {
    influe.actions['request'] = async (context, argument) => {
        const uri = argument.uri || argument.url;
        const outPath = argument.out;
        if (argument.out === undefined) {
            throw new Error('argument.out is not specified.');
        }
        influe.os.createImmidateDirectories(outPath);
        const options = {
            method: argument.method || 'GET',
            ...argument.options,
        };
        requestHttp(uri, options)
            .then(result => {
            const { response, body } = result;
            fs.writeFileSync(outPath, body, { encoding: 'utf8' });
        })
            .catch(error => {
            throw new Error(error);
        });
    };
    async function requestHttp(uri, options) {
        return new Promise((resolve, reject) => {
            httpRequest(uri, options, (error, response, body) => {
                if (error) {
                    reject(error);
                }
                else {
                    resolve({ response, body });
                }
            });
        });
    }
})(influe || (influe = {}));
(function (influe) {
    influe.actions['mail'] = async (context, argument) => {
        const host = context.data.default.smtpHost || '';
        const port = context.data.default.smtpPort || 587;
        const user = context.data.default.smtpUser || '';
        const pass = context.data.default.smtpPassword || '';
        // create reusable transporter object using the default SMTP transport
        const transporter = nodemailer.createTransport({
            host,
            port,
            secure: port == 465,
            auth: {
                user,
                pass,
            }
        });
        // send mail with defined transport object
        const info = await transporter.sendMail({
            from: argument.from,
            to: influe.wrapInArray(argument.to),
            cc: influe.wrapInArray(argument.cc),
            bcc: influe.wrapInArray(argument.bcc),
            subject: argument.subject,
            ...normalizeBody(argument.body),
        });
        console.log("Message sent: %s", info.messageId);
        // Message sent: <b658f8ca-6296-ccf4-8306-87d57a0b4321@example.com>
    };
    function normalizeBody(argument) {
        let body;
        if (Array.isArray(argument)) {
            body = { text: argument };
        }
        else if (typeof argument == 'object') {
            body = argument;
        }
        else {
            body = { text: argument };
        }
        for (const key in body) {
            body[key] = influe.wrapInArray(body[key]).join('\r\n');
        }
        return body;
    }
})(influe || (influe = {}));
(function (influe) {
    influe.actions['yarn'] = async (context, argument) => {
        const args = influe.wrapInArray(argument.command)
            .map(line => influe.expandTemplateText(line, context, process.env));
        const cwd = argument.cwd;
        influe.pauseConsoleWatch();
        try {
            await influe.os.executeCommand('yarn', args, {
                cwd: cwd
            });
        }
        catch (error) {
            console.error(error);
        }
        influe.resumeConsoleWatch();
    };
})(influe || (influe = {}));
(function (influe) {
    var os;
    (function (os) {
        // process.on('SIGINT', function () {
        // 	console.log('SIGINT')
        // })
        // console.log(90)
        function readTextFile(path) {
            return fs.readFileSync(path, { encoding: 'utf8' });
        }
        os.readTextFile = readTextFile;
        function writeTextFile(path, content) {
            createImmidateDirectories(path);
            fs.writeFileSync(path, content, { encoding: 'utf8' });
        }
        os.writeTextFile = writeTextFile;
        function createImmidateDirectories(path) {
            const dirname = fspath.dirname(path);
            if (!fs.existsSync(dirname)) {
                fs.mkdirSync(dirname, { recursive: true });
            }
        }
        os.createImmidateDirectories = createImmidateDirectories;
        function unlinkFile(path) {
            if (fs.existsSync(path)) {
                fs.unlinkSync(path);
            }
        }
        os.unlinkFile = unlinkFile;
        function executeShell(command, options = {}) {
            options.env = process.env;
            exec(command, options, (error, stdout, stderr) => {
                if (error) {
                    console.log(chalk.red('Error:' + error));
                }
                if (stdout) {
                    process.stdout.write(stdout);
                }
                if (stderr) {
                    process.stdout.write(stderr);
                }
            });
        }
        os.executeShell = executeShell;
        async function executeCommand(command, args, options = {}) {
            return new Promise((resolve, reject) => {
                const rawMode = options.rawMode ?? false;
                const child = spawn(command, args, { ...options, stdio: rawMode ? 'inherit' : ['pipe', 'inherit', 'inherit'] });
                if (rawMode == false) {
                    install();
                }
                child.on('close', (exitCode, signal) => {
                    console.log("close", exitCode, signal);
                    resolve(exitCode);
                });
                child.on('error', error => {
                    console.error("error");
                    reject(error);
                });
                function install() {
                    process.stdin.setRawMode(true);
                    process.stdin.on('data', onKeypress);
                }
                function uninstall() {
                    process.stdin.setRawMode(false);
                    process.stdin.off('data', onKeypress);
                }
                function onKeypress(key) {
                    const keyName = keyCodeToName(key);
                    if (key === KEYS.CONTROL_C) {
                        console.log("CTRL+C", keyName);
                        child.kill('SIGINT');
                        child.kill('SIGTERM');
                        uninstall();
                    }
                    else if (key === KEYS.CONTROL_D) {
                        console.log("CTRL+D", keyName);
                        child.kill('SIGTERM');
                        uninstall();
                    }
                    else {
                        if (key === KEYS.ENTER) {
                            process.stdout.write(key);
                            child.stdin?.write("\r\n");
                        }
                        else {
                            child.stdin?.write(key);
                        }
                    }
                }
            });
        }
        os.executeCommand = executeCommand;
    })(os = influe.os || (influe.os = {}));
})(influe || (influe = {}));
(function (influe) {
    let DataReader;
    (function (DataReader) {
        const readers = {
            '.json': readJson,
        };
        function isSupportFile(filePath) {
            const extension = fspath.extname(filePath).toLowerCase();
            return Object.keys(readers).includes(extension);
        }
        DataReader.isSupportFile = isSupportFile;
        function read(filePath) {
            const extension = fspath.extname(filePath).toLowerCase();
            const reader = readers[extension] || null;
            if (!reader)
                throw new Error('Unknown data file extension.');
            return reader(filePath);
        }
        DataReader.read = read;
        function readJson(filePath) {
            return JSON.parse(influe.os.readTextFile(filePath));
        }
        DataReader.readJson = readJson;
    })(DataReader = influe.DataReader || (influe.DataReader = {}));
})(influe || (influe = {}));
(function (influe) {
    function expandTemplateText(text, context, locals) {
        return text.replace(/\$\{(.*?)\}/msg, ($0, $1) => {
            return evaluateExpression(context, locals, $1, '') || $0;
        });
    }
    influe.expandTemplateText = expandTemplateText;
    function evaluateExpression(context, locals, script, location = '') {
        const scaned = scriptkit.scan({ input: script, location }, 'scriptkit/expression');
        if (scaned == null) {
            return null;
        }
        const world = scriptkit.createWorld([
            { id: '^', thing: globals },
            { id: '~', thing: context },
        ], []);
        return scriptkit.evaluate(world, locals, scaned);
    }
    influe.evaluateExpression = evaluateExpression;
    const globals = {
        env: process.env,
    };
})(influe || (influe = {}));
(function (influe) {
    function runAll(config, taskIds) {
        for (const taskId of taskIds) {
            const task = config.tasks.find(_ => _.id == taskId);
            if (!task)
                throw new Error('task not found.');
            runTask(task.id, task, { data: config.data });
        }
    }
    influe.runAll = runAll;
    function runTask(name, task, context) {
        influe.touch(influe.TouchPoint.Task, 'pre', chalk.grey('name:') + chalk.underline(name), chalk.grey('id:') + chalk.underline(task.id), chalk.grey('action:') + chalk.underline(task.action));
        context.task = {
            id: task.id,
            name,
            action: task.action,
        };
        if (task.action in influe.actions) {
            try {
                influe.runAction(task.action, task.arguments, context);
            }
            catch (e) {
                influe.touch(influe.TouchPoint.Task, 'error', e);
            }
        }
        else {
            throw new Error(`action ${task.action} is not found.`);
        }
    }
    influe.runTask = runTask;
    function makeAll(config, targets) {
        for (const target of targets) {
            if (target.files.length == 0) {
                makeState(config, target.state);
            }
            else {
                for (const file of target.files) {
                    makeOneState(config, file, target.state);
                }
            }
        }
    }
    influe.makeAll = makeAll;
    function makeState(config, state) {
        const targetFiles = config.files.filter(_ => Object.keys(_.states).includes(state));
        if (targetFiles.length == 0) {
            console.error(`State "${state}" is not found.`);
        }
        else {
            targetFiles.map(file => {
                updateFile(config, file, state);
            });
        }
    }
    function makeOneState(config, fileId, state) {
        const file = config.files.find(_ => _.id == fileId);
        if (file === undefined)
            throw new Error(`File "${fileId}" is not found.`);
        if (state in file.states) {
            updateFile(config, file, state);
        }
        else {
            throw new Error(`State "${state}" is not found.`);
        }
    }
    function updateFile(config, file, state) {
        const name = `${file.id}.${state}`;
        for (const taskOrRef of file.states[state]) {
            // taskRef : string
            if (typeof taskOrRef === 'string') {
                const task = config.tasks.find(_ => _.id == taskOrRef);
                if (task === undefined) {
                    throw new Error(`Task "${taskOrRef}" is not found.`);
                }
                runTask(taskOrRef, task, { data: config.data });
            }
            // task : Task
            else {
                runTask(name, taskOrRef, { data: config.data, file });
            }
        }
    }
    influe.updateFile = updateFile;
})(influe || (influe = {}));
(function (influe) {
    const jobs = [];
    function startTimeWatch(config) {
        const context = { data: config.data };
        for (const _ of config.events.filter(_ => _.type == 'time:cron')) {
            const event = _;
            const crontime = event.crontime;
            const timezone = influe.expandTemplateText(event.timezone, context, {}) || config.data?.default?.timezone || 'UTC';
            influe.touch(influe.TouchPoint.Watch, 'log', 'Prepare:' + chalk.green(event.id), chalk.cyan('Time'), event.enabled ? '' : chalk.red('[DISABLED]'), `"${crontime}" ${timezone}`);
            if (event.enabled == false) {
                continue;
            }
            launchCronJob(crontime, timezone, () => influe.runAll(config, event.runTargets));
        }
        // for (const _ of config.events.filter(_ => _.type == 'time:xxx')) {
        //     const event = _ as TimeEvent_
        //     startCronJob(config, event.crontime, event.timezone, event.runTargets)
        // }
    }
    influe.startTimeWatch = startTimeWatch;
    function launchCronJob(crontime, timezone, onTick) {
        const job = cron.job(crontime, onTick, undefined, false, timezone);
        job.start();
        jobs.push(job);
    }
    function stopTimeWatch() {
        for (const job of jobs) {
            job.stop();
        }
        jobs.length = 0;
    }
    influe.stopTimeWatch = stopTimeWatch;
})(influe || (influe = {}));
(function (influe) {
    const watchers = [];
    function startFileWatch(config) {
        const context = { data: config.data };
        for (const _ of config.events.filter(_ => _.type == 'file')) {
            const event = _;
            influe.touch(influe.TouchPoint.Watch, 'log', 'Prepare:' + chalk.green(event.id), chalk.cyan('File'), event.enabled ? '' : chalk.red('[DISABLED]'), `"${event.paths.join(',')}" on (${event.triggers.join('|')})`);
            if (event.enabled == false) {
                continue;
            }
            launchFileMonitor(event.paths, (eventName, path) => {
                if (config.debug)
                    console.log(`"${eventName}" : "${path}"`);
                if (event.triggers.some(_ => _ == eventName)) {
                    try {
                        influe.makeAll(config, event.makeTargets);
                        influe.runAll(config, event.runTargets);
                    }
                    catch (error) {
                        influe.touch(influe.TouchPoint.Watch, 'error', error);
                    }
                }
            });
        }
    }
    influe.startFileWatch = startFileWatch;
    function launchFileMonitor(paths, onNotify) {
        const watcher = chokidar.watch(paths, {
            cwd: '.',
            followSymlinks: true,
            ignoreInitial: true,
            persistent: true,
        });
        watcher.on('all', onNotify);
        watchers.push(watcher);
    }
    function stopFileWatch() {
        for (const watcher of watchers) {
            watcher.close();
        }
        watchers.length = 0;
    }
    influe.stopFileWatch = stopFileWatch;
})(influe || (influe = {}));
const isWindows = process.platform === 'win32';
const KEYS = {
    ARROW_UP: '\u001b[A',
    ARROW_DOWN: '\u001b[B',
    ARROW_RIGHT: '\u001b[C',
    ARROW_LEFT: '\u001b[D',
    TAB: '\t',
    ENTER: '\r',
    ESCAPE: '\u001b',
    SPACE: ' ',
    BACKSPACE: Buffer.from(isWindows ? '08' : '7f', 'hex').toString(),
    CONTROL_C: '\u0003',
    CONTROL_D: '\u0004',
    CONTROL_U: '\u0015',
};
const KEY_MAP = new Map([
    ['up', KEYS.ARROW_UP],
    ['down', KEYS.ARROW_DOWN],
    ['right', KEYS.ARROW_RIGHT],
    ['left', KEYS.ARROW_LEFT],
    ['tab', KEYS.TAB],
    ['enter', KEYS.ENTER],
    ['escape', KEYS.ESCAPE],
    ['space', KEYS.SPACE],
    ['backspace', KEYS.BACKSPACE],
    ['delete', '\u001b[3~'],
]);
function keyNameToCode(name) {
    const code = KEY_MAP.get(name);
    return code ? code : name;
}
function keyCodeToName(code) {
    for (const [key, value] of KEY_MAP) {
        if (code == value) {
            return key;
        }
    }
    return code;
}
(function (influe) {
    // REF: https://github.com/facebook/jest/tree/main/packages/jest-core/src/watch.ts
    const keyActions = [];
    const stdin = process.stdin;
    let initialized = false;
    async function startConsoleWatch(config) {
        if (initialized == false) {
            installConsoleWatch();
            initialized = true;
        }
        // 構成のロード
        for (const _ of config.events.filter(_ => _.type == 'key')) {
            const event = _;
            influe.touch(influe.TouchPoint.Watch, 'log', 'Prepare:' + chalk.green(event.id), chalk.cyan('Key'), event.enabled ? '' : chalk.red('[DISABLED]'), `"${event.key}"`, '=>', `${event.runTargets.join(', ')}`);
            if (event.enabled == false) {
                continue;
            }
            // TODO: event.keyをパースして、KeyStrokeに変換する
            const key = {
                key: event.key,
                control: false,
            };
            keyActions.push({
                keys: [key],
                onPressed: () => {
                    try {
                        influe.makeAll(config, event.makeTargets);
                        influe.runAll(config, event.runTargets);
                    }
                    catch (error) {
                        influe.touch(influe.TouchPoint.Watch, 'error', error);
                    }
                }
            });
        }
    }
    influe.startConsoleWatch = startConsoleWatch;
    function installConsoleWatch() {
        // キー入力イベントのリスニング
        // NOTE: 行入力モードからダイレクト入力モードに切り替える
        stdin.setRawMode(true);
        stdin.resume();
        stdin.setEncoding('utf8');
        stdin.on('data', onKeypress);
    }
    function uninstallConsoleWatch() {
        stdin.setRawMode(false);
        stdin.resume();
        stdin.off('data', onKeypress);
    }
    function onKeypress(key) {
        if (key === KEYS.CONTROL_C || key === KEYS.CONTROL_D) {
            if (typeof stdin.setRawMode === 'function') {
                stdin.setRawMode(false);
            }
            influe.command.stopWatchAndExit();
            return;
        }
        const keyName = keyCodeToName(key);
        influe.touch(influe.TouchPoint.Watch, 'log', chalk.cyan('Key'), `"${keyName}" ${key.length > 2 ? '(ESC+' + key.substring(1) + ')' : ''}`);
        match(keyName)?.onPressed();
    }
    function match(key) {
        return keyActions.find(_ => _.keys[0].key == key);
    }
    function stopConsoleWatch() {
        keyActions.length = 0;
    }
    influe.stopConsoleWatch = stopConsoleWatch;
    function pauseConsoleWatch() {
        if (initialized == false)
            return;
        uninstallConsoleWatch();
    }
    influe.pauseConsoleWatch = pauseConsoleWatch;
    function resumeConsoleWatch() {
        if (initialized == false)
            return;
        installConsoleWatch();
    }
    influe.resumeConsoleWatch = resumeConsoleWatch;
})(influe || (influe = {}));
export default influe;
//# sourceMappingURL=influe.mjs.map